#!/usr/bin/env python

"""From davem@magnet.com Wed Nov 26 16:21:26 1997
Date: Tue, 25 Nov 1997 15:17:02 -0500 (EST)
From: Dave Mitchell <davem@magnet.com>
To: python-list@cwi.nl
Subject: python/tkinter othello games



  Here's something I wrote the last couple of evenings - a Tkinter
othello game. Although it's a two-player game, I havent provided for
either multi-player or computer player use. I don't expect anyone will,
but two people would have to share the same mouse to play each other. 
I just like watching the neat shapes made by the "possible moves" 
rectangles ;)

							-dave

 ps- I do plan to do one or both of the above things I mentioned i
hadn't done yet. Maybe soon. maybe not. I think it's just a neat 
tkinter demo. Plus it runs under the python1.4 windows binary I
downloaded from http://www.pythonware.com/downloads.htm (if a 
little slowly..)


Copyright 1997, David W. Mitchell (davem@magnet.com)

version 0.1 11/25/97

anyone is free to use this code as they see fit, but I take
no responsibility if you do something stupid like using it to 
launch nukes and they go off course and blow up the wrong people.

this program implements the game othello. there is no computer
player just yet, and it's basically just a vehicle I used to
learn more about Canvases and tags under Tkinter in Python. Next things
I want to do for this are 1) a computer player, and 2) a multiplayer
version using either ilu or some plain old homegrown socket stuff
(since I can't get ilu working on my non-networked home machine for
some reason)

to run it, just do "python othello.py". I got it to work for me
under Linux 2.0.30, Irix 6.2, and Win'95.

A window will appear containing the game board, the names "white" and 
"black" below the board, and a "Start Game" button.

Press "Start Game" to begin. The first player's name will be highlighted
in red, and a series of small red (and one blue) squares will be drawn
on the board. The squares represent the locations where it is legal
for the current player to move. The blue location is the one which 
will yield the most pieces.

When the current player has no moves left, the game is over and the
winner is the one with the most pieces. Click the "Quit" button to
end the program.

"""

from Tkinter import *

class Location:

# each of the 64 spaces on the 8x8 othello board have a Location 
# instance associated with it. 

  def __init__(self, parent, x, y):
    self.parent = parent
    self.color = None
    self.x = x
    self.y = y

class TurnIndicator(Frame):

# this class maintains list of the players in the game, and
# returns a reference to the current player or increments the
# index and returns the next player. It also creates a visual
# display of the players and changes the background of the
# "current" player's name to red.

  def __init__(self, parent, players, **kw):
    Frame.__init__(self, parent, kw)
    self.players = players
    self.player_frames = []
    self.index = -1
    for player in players:
      tmp = Label(self,text="%s"%(player),
                background="grey",foreground=player,
                relief=RAISED)
      self.player_frames.append(tmp)
      tmp.pack(side=LEFT,ipadx=5)

  def next(self):
    """increment self.index to the next player, highlight the current player"""
    if self.index > -1:
      bg = "grey"
      self.player_frames[self.index].config(\
                background=bg)
    self.index = self.index + 1
    if self.index >= len(self.players):
      self.index = 0
    bg = "red" 
    self.player_frames[self.index].config(\
                background=bg)
    return self.players[self.index]

  def current(self):
    if self.index > -1:
      return self.players[self.index]
 
class Othello(Canvas):
  """This is the big mama-jama class that handles the whole playing area.

  This contains mostly:

  self.board, a list of all the rows on the board.
    i.e. location x,y is referenced via "self.board[y][x]" 

  self.players, a list of the two players (just "white" and "black")
 
  self.pieces, a dictionary, keyed by the player names "white" and "black",
	with a list of all of those player's pieces.

  self.outcomes, a dictionary, created by mark_possible_moves(), that
    	contains, for each turn, a mapping between possibly-played
	locations on the board and the locations which would be turned
        over if the player moved to that location. This also allows
	me to to easily figure which move yields the highest number
        of points. 

  The class calls self.boardinit() in it's __init__() method. This 
	initializes the array of Location instances (self.board),
	and for each, draws a square on the playing field representing
	itself. The square gets a tag like "2,3" which contains the
	square's location on the board. The same tag is also bound
	to "<Button-1>" to call (what I guess is a form of closure)
	a lambda function whose arguments are pre-filled at this time.
	In this way, I can map the user's clicks to the appropriate
	Location instance. This is useful since I can't actually add
	tags to instances. I can only add tags to the onscreen shapes
	representing each Location instance. Get it? ;)

   self.neighbors() takes an x and y coordinate for a board location,
	and an optional color to specify returning only neighbors of
	that color. If the color is given as None, then all neighbors
	with -no- playing piece will be returned. The return value is
	a dictionary, whose keys are 2-tuples indicating the x and y
	offset from the center location, and whose value is the 
	Location reference of the particular neighbor. The offset tuple
	is required in order to loop down consecutive adjacent neighbors 
        in the same direction to determine legal moves.

   self.mark_possible_moves() takes the color of the player whose turn
	it is. It starts by building a list of all unplayed neighbors
	of all the opposing player's pieces. This is called "initial".
	For each location in "initial", a loop travels outward in each
	of the 8 possible directions, checking that they consist of
	one or more of the opposite player's pieces, and then at least
	one of the current player's pieces. All the locations that pass
	this second test are saved into a list called "final". The 
	paths are saved into self.outcomes during this process as well.
	Finally, for each location in "final", it draws small red 
	rectangles on the screen, each containing tags for the location
	(x,y) and "temp" indicating that it can be deleted next turn.
        The outcomes dictionary's values are sorted by length, and the
	possible move that would yield the greatest number of points
	is marked in blue. If there are more than one move tied for 
	"best", only the first is shown.

  """
  
  def __init__(self, parent=None, **kw):
    Canvas.__init__(self, parent, kw)
    self.config(background="darkgreen") 
    self.parent=parent
    self.pack()
    self.size = 8                      # 8 across x 8 down
    self.players = []
					# dict to access player's pieces
    self.outcomes = {}
    self.boardinit()

  def join(self, player):
    self.players.append(player)

  def get_loc(self, x, y):
    return self.board[y][x]

  def boardinit(self):
    self.board = []
    self.pieces = {"black":[], "white":[]}
    w,h = self.getint(self['width']),self.getint(self['height'])
    self.xinc,self.yinc = w/self.size, h/self.size
    self.piece_width = self.xinc*0.75
    self.hw = self.piece_width*0.5
    self.xih,self.yih = self.xinc*0.5,self.yinc*0.5

    x,y = 0,0
    for row in range(self.size):
      self.board.append([])
      for col in range(self.size):
        loc = self.create_rectangle(x,y,x+self.xinc,y+self.yinc,\
		fill="darkgreen",\
		tags="%d,%d"%(col,row))
        self.tag_bind("%d,%d"%(col,row),"<Button-1>",\
		lambda x,r=row,c=col,s=self: s.pressed(c,r))
        tmp = Location(self, col, row)
        self.board[row].append(tmp)
        x = x + self.xinc
      y = y + self.yinc
      x = 0

    self.go_loc(3,3,"white")
    self.go_loc(4,4,"white")
    self.go_loc(3,4,"black")
    self.go_loc(4,3,"black")	# set out the initial pieces
    self.delete('temp')  	# unhighlight the moves before the game starts
    self.dtag("selectable","selectable")

  def pressed(self, x, y):
    current = turn.current()
    if current and current != self.board[y][x].color:
      t = self.find_withtag("%d,%d"%(x,y))
      if "selectable" in self.gettags(t[0]):
        self.go_loc(x,y,current)
        self.next()

  def go_loc(self, x, y, color, recur=1):
    loc = self.board[y][x]       
    if loc.color:
      self.pieces[loc.color].remove(loc)
    loc.color = color
    x1 = x*self.xinc + self.xih
    y1 = y*self.yinc + self.yih
    self.create_oval(x1-self.hw, y1-self.hw,
		x1+self.hw, y1+self.hw, fill=color,
		tags="%d,%d"%(loc.x,loc.y))
    self.pieces[color].append(loc)	
    if recur:
      try:
        for l in self.outcomes[loc]:
          self.go_loc(l.x,l.y,color,0)
      except KeyError:
        pass

  def neighbors(self, loc, color="any"):

    n = {}
    for xoff,yoff in ((-1,-1),(0,-1),(1,-1),(-1,0),
                        (1,0),(1,1),(0,1),(-1,1)):
      nx = loc.x+xoff
      ny = loc.y+yoff
      if nx >= 0 and nx < self.size and ny >= 0 and ny < self.size:
        p = self.board[ny][nx]
        if color =="any" or p.color == color:
          n[(xoff,yoff)] = self.board[ny][nx]
    return n

  def mark_possible_moves(self, color):
    outcomes = {}
    self.delete("temp")
    self.dtag("selectable","selectable")
    initial = []
    if color == "white":
      other = "black"
    elif color == "black":
      other = "white"
    enemy_owned = self.pieces[other]
    for loc in enemy_owned:
      neighbors = self.neighbors(loc,None)
      for n in neighbors.values():
        if n not in initial:
          initial.append(n)

# this the first cut at all the legal moves
# now we have to remove all locations where
# there is no chain of enemy pieces between
# the location and another of the player's
# pieces.. 

    final = []

    for loc in initial:
      n = self.neighbors(loc, other)
      for k,v in n.items():
        path = []
	while v.color == other and v.x+k[0] >= 0 and v.x+k[0] < self.size \
                and v.y+k[1] >= 0 and v.y+k[1] < self.size:
          path.append(v)
          v = self.board[v.y+k[1]][v.x+k[0]]
        if v.color == color:
          try:
            outcomes[loc] = outcomes[loc] + path
          except KeyError:
            outcomes[loc] = path
          if loc not in final:
            final.append(loc)
        else:
            pass

# at this point outcomes contains all the paths built in the above
# loop, and final contains the list of validated "initial" locations.
 
    self.outcomes = outcomes
    
    lst = outcomes.items()
    lst.sort(lambda a,b: len(a[1]) < len(b[1]))
    try:
      best_move = lst[0][0]
    except:
      best_move = ""

# sorting the item() list gives the location which will yield
# the greatest number of pieces.

    if not final:
      self.gameover()

    for loc in final:
      if loc is best_move:
        fillcolor="blue"
      else:
        fillcolor="red"
      tmp = self.create_rectangle(loc.x*self.xinc+self.xih-5, \
			    loc.y*self.yinc+self.yih-5, \
			    loc.x*self.xinc+self.xih+5, \
			    loc.y*self.yinc+self.yih+5,
		            tags="temp", fill=fillcolor)
      self.addtag_withtag("%d,%d"%(loc.x,loc.y), tmp)
      self.addtag_withtag("selectable", "%d,%d"%(loc.x,loc.y))

# only "selectable" objects are able to be played on 

  def gameover(self):
    a,b=len(self.pieces['black']),len(self.pieces['white'])
    if len(self.pieces['black']) > len(self.pieces['white']):
      winner = "black"
      loser = "white"
    else:
      winner = "white"
      loser="black"
      a,b=b,a

    w,h = self.getint(self['width']),self.getint(self['height'])
    self.create_text(w/2,h/2,font="12x24",text="Game Over",fill="red",\
	tags="winmsg")
    self.winmsg = Label(self.parent,font="*-*-14",\
	text="%s beats %s, %d to %d!"%(winner,loser,a,b))
    self.winmsg.pack()
    start_textvar.set("Quit")
    start_button.config(command=start_button.root.destroy)
    start_button.pack()

  def next(self, event=None):
    global turn
    next = turn.next()
    self.mark_possible_moves(next)

def start(othref):
  global start_button, turn
  othref.next()
  start_button.forget()

turn = ""
start_textvar = ""
start_button = ""

if __name__ == "__main__":
  global turn, start_textvar, start_button
  root = Tk()
  x=Othello(root,width=300,height=300)
  root.title("Othello")
  players = ["black", "white"]
  x.join(players[0])
  x.join(players[1])
  turn = TurnIndicator(root,players)
  turn.pack()
  f = Frame(root)
  f.pack()
  start_textvar = StringVar()
  start_textvar.set("Start Game")
  start_button = Button(f,textvariable=start_textvar,\
			command=lambda x=x,s=start:s(x))
  start_button.root = root
  start_button.pack()
  root.mainloop()
 



